const fs = require('fs');

const DECIMALS = 4;
const ROTATING_ANGLE = 5;
const SYMMETRIC_LIMIT_ANGLE = 120;

function getFilePathForCounter(counter) {
  return './src/maps/' + counter + '.map';
}

function readBoardData() {
  let data = fs.readFileSync(getFilePathForCounter(1), 'utf8');
  // parse JSON string to JSON object
  return JSON.parse(data);
}

function rollup(pos) {
  return {
    x: parseFloat(pos.x.toFixed(DECIMALS)),
    y: pos.y,
    z: parseFloat(pos.z.toFixed(DECIMALS)),
  };
}

function transform(pos, deg) {
  let angle = (deg * Math.PI) / 180;
  return {
    x: pos.x * Math.cos(angle) + pos.z * Math.sin(angle),
    y: pos.y,
    z: pos.z * Math.cos(angle) - pos.x * Math.sin(angle),
  };
}

const deepClone = obj => {
  if (obj === null) return null;
  let clone = Object.assign({}, obj);
  Object.keys(clone).forEach(
      key =>
          (clone[key] =
              typeof obj[key] === 'object' ? deepClone(obj[key]) : obj[key]),
  );
  return Array.isArray(obj) && obj.length
      ? (clone.length = obj.length) && Array.from(clone)
      : Array.isArray(obj)
          ? Array.from(obj)
          : clone;
};

let actualBoardData;

function getCopyData() {
  if (!actualBoardData) actualBoardData = readBoardData();
  return deepClone(actualBoardData);
}

function writeForADegree(deg, counter) {
  let board = getCopyData();
  for (let go of board.gameObjects) {
    if (!['wcoin', 'bcoin'].includes(go.type)) {
      continue;
    }
    go.position = rollup(transform(go.position, deg));
  }
  fs.writeFileSync(getFilePathForCounter(counter),
      JSON.stringify(board, null, 2));
}

const exec = async function() {
  let counter = 1;
  for (let i = ROTATING_ANGLE; i < SYMMETRIC_LIMIT_ANGLE; i += ROTATING_ANGLE)
    writeForADegree(i, ++counter);
};

exec().catch(console.error).finally(process.exit);