const oddUtils = require('../utils/oddUtils');
const logger = oddUtils.getLogger();
const turnFillUtils = require('../utils/fillAndFetchUtils').turnFillUtils;
const {_getGameProp,_setGameProp,_getPlayerProp,_setPlayerProp} =
    require('../utils/utils')
const ENUMS = require("../utils/enums")
const OBJ_TYPES = oddUtils.listOfTypes();
const ALL_PROP_KEYS = oddUtils.getAllPropKeys();

function collisionHandler(game, playerId, objIdA, objIdB) {
  const gameKey = game.getId();

  if (game.getCurrentTurnPlayerId() !== playerId)
    return logger.warn({key: gameKey},
        'Collision report received out of turn, dropping it');
  let objA = game.gameObjectByIds(objIdA)[objIdA];
  let objB = game.gameObjectByIds(objIdB)[objIdB];
  let isStrikerWhiteCoinCollision = collisionTypeDetector(objA, objB,
      [OBJ_TYPES.STRIKER, OBJ_TYPES.WHITE_COIN]);
  let isStrikerBlackCoinCollision = collisionTypeDetector(objA, objB,
      [OBJ_TYPES.STRIKER, OBJ_TYPES.BLACK_COIN]);
  logger.debug({key: gameKey},"striker first collision with white and black",
      isStrikerWhiteCoinCollision,isStrikerBlackCoinCollision)
  if(isStrikerWhiteCoinCollision || isStrikerBlackCoinCollision ){
    if(game.getGameProp("isBreak") === 'true') game.setGameProp("isBreak", false);
  }
  let isStrikerHoleCollision = collisionTypeDetector(objA, objB,
      [OBJ_TYPES.STRIKER, OBJ_TYPES.HOLE]);
  if (isStrikerHoleCollision) {
    handleStrikerHoleCollision(game, playerId, objA, objB);
    return;
  }

  let isCoinHoleCollision =
      collisionTypeDetector(objA, objB, [OBJ_TYPES.RED_COIN, OBJ_TYPES.HOLE])
      ||
      collisionTypeDetector(objA, objB, [OBJ_TYPES.BLACK_COIN, OBJ_TYPES.HOLE])
      ||
      collisionTypeDetector(objA, objB, [OBJ_TYPES.WHITE_COIN, OBJ_TYPES.HOLE]);

  if (isCoinHoleCollision) {
    handleCoinHoleCollision(game, playerId, objA, objB);
    logger.debug({key: gameKey}, 'Handling collision between ', objA.type, objB.type, 'for ',
        playerId);
    return;
  }

  logger.trace({key: gameKey}, 'Ignoring the collision between ', objA.type, objB.type, 'for ',
      playerId);
}

function handleCoinHoleCollision(game, playerId, objA, objB) {
  const gameKey = game.getId();
  // find which obj is coin
  let coin = objA.type === OBJ_TYPES.HOLE ? objB : objA;
  // ignore already scored coins collisions
  if (oddUtils.parseBoolean(coin.getProp(ALL_PROP_KEYS.IS_SCORED))) {
    logger.trace({key: gameKey}, 'Coin is already scored ignoring');
    return;
  }
  _setPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.PLAYER_TURN_POCKETED_COUNT,
      _getPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.PLAYER_TURN_POCKETED_COUNT)+1)
  // mark coin as potted
  coin.setProp(ALL_PROP_KEYS.IS_POTTED, true);
  // mark coin for current tack and current turn
  coin.setProp(ALL_PROP_KEYS.POTT_TURN, game.getCurrentTurnIndex());
  coin.setProp(ALL_PROP_KEYS.POTT_TACK,
      game.getGameProp(ALL_PROP_KEYS.POTT_TACK));

  // push coin into temp mem fill
  turnFillUtils.pushPottedCoinToCache(game, coin, game.getCurrentTurnIndex(),
      game.getGameProp(ALL_PROP_KEYS.POTT_TACK));
}

function handleStrikerHoleCollision(game, playerId, objA, objB) {
  // find which obj is striker
  _setPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.STRIKER_POCKETED_COUNT,
      _getPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.STRIKER_POCKETED_COUNT)+1)
  let striker = objA.type === OBJ_TYPES.STRIKER ? objA : objB;
  // mark striker as potted
  striker.setProp(ALL_PROP_KEYS.IS_POTTED, true);
}

function collisionTypeDetector(objA, objB, detectionPair) {
  return (objA.type === detectionPair[0] && objB.type === detectionPair[1]) ||
      (objA.type === detectionPair[1] && objB.type === detectionPair[0]);
}

module.exports = {collisionHandler};

