const oddUtils = require('../utils/oddUtils');
const logger = oddUtils.getLogger();

const OBJ_TYPES = oddUtils.listOfTypes();
const ALL_PROP_KEYS = oddUtils.getAllPropKeys();

module.exports.turnFillUtils = {
  initEmptyTurnCache: function(game, turnIndex) {
    if (!game.fillFetch) game.fillFetch = {};
    if (!game.fillFetch.coinPot) game.fillFetch.coinPot = {};
    let coinPot = game.fillFetch.coinPot;
    coinPot[turnIndex] = oddUtils.FILL_FETCH_DEFAULT;
  },

  initEmptyTackCache: function(game, turnIndex, tack) {
    if (!game.fillFetch || !game.fillFetch.coinPot ||
        !game.fillFetch.coinPot[turnIndex]) return;
    if (game.fillFetch.coinPot[turnIndex] === oddUtils.FILL_FETCH_DEFAULT)
      game.fillFetch.coinPot[turnIndex] = {};
    game.fillFetch.coinPot[turnIndex][tack] = oddUtils.FILL_FETCH_DEFAULT;
  },

  pushPottedCoinToCache: function(game, coin, turnIndex, tack) {
    if (!game.fillFetch || !game.fillFetch.coinPot ||
        !game.fillFetch.coinPot[turnIndex] ||
        !game.fillFetch.coinPot[turnIndex][tack])
      return;
    if (game.fillFetch.coinPot[turnIndex][tack] === oddUtils.FILL_FETCH_DEFAULT)
      game.fillFetch.coinPot[turnIndex][tack] = [];
    game.fillFetch.coinPot[turnIndex][tack].push(coin);
  },
};

module.exports.turnFetchUtils = {
  fetchCoinsForTurnAndTack: function(game, turnIndex, tack) {

    // check if value is stored in fillFetch cache
    // if (game.fillFetch && game.fillFetch.coinPot &&
    //     game.fillFetch.coinPot[turnIndex] &&
    //     game.fillFetch.coinPot[turnIndex][tack]) {
    //   if (game.fillFetch.coinPot[turnIndex][tack] ===
    //       oddUtils.FILL_FETCH_DEFAULT) return []; // No coin has been potted
    //   // values are stored in cache so fetch from it
    //   return game.fillFetch.coinPot[turnIndex][tack];
    // }
    // if not present in cache, then iterate over coins and fetch the list
    turnIndex = parseInt(turnIndex);
    tack = parseInt(tack);
    // get all coins
    let coins = [
      ...fetchObjsByType(game, OBJ_TYPES.BLACK_COIN),
      ...fetchObjsByType(game, OBJ_TYPES.WHITE_COIN),
      ...fetchObjsByType(game, OBJ_TYPES.RED_COIN)];
    let pottedList = [];
    for (let coin of coins) {
      let isPotted = oddUtils.parseBoolean(
          coin.getProp(ALL_PROP_KEYS.IS_POTTED));
      let isScored = oddUtils.parseBoolean(
          coin.getProp(ALL_PROP_KEYS.IS_SCORED));

      if (isPotted && !isScored) {
        let coinTurnIndex = parseInt(coin.getProp(ALL_PROP_KEYS.POTT_TURN));
        let coinTack = parseInt(coin.getProp(ALL_PROP_KEYS.POTT_TACK));
        if (turnIndex === coinTurnIndex && tack === coinTack)
          pottedList.push(coin);
      }
    }
    return pottedList;
  },
};

function fetchObjsByType(game, type) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.objTypeMap) game.fillFetch.objTypeMap = {};
  if (game.fillFetch.objTypeMap[type])
    return game.fillFetch.objTypeMap[type];
  let list = game.fillFetch.objTypeMap[type] = [];
  let objIds = game.getAllObjIds();
  let allObjs = game.gameObjectByIds(...objIds);
  for (let objId of objIds) {
    let obj = allObjs[objId];
    if (obj.type === type)
      list.push(obj);
  }
  return list;
}

function fetchObjById(game, id) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.objs) game.fillFetch.objs = {};
  if (!game.fillFetch.objs[id])
    game.fillFetch.objs[id] = game.gameObjectByIds(id)[id];
  return game.fillFetch.objs[id];
}

const fetchStriker = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.striker) {
    let list = fetchObjsByType(game, OBJ_TYPES.STRIKER);
    if (list.length !== 1) throw new Error('Multiple striker objects found.'); // Something seriously went wrong
    game.fillFetch.striker = list[0];
  }
  return game.fillFetch.striker;
};

const fetchHand = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.hand) {
    let list = fetchObjsByType(game, OBJ_TYPES.HAND);
    if (list.length !== 1) throw new Error('Multiple hand objects found.'); // Something seriously went wrong
    game.fillFetch.hand = list[0];
  }
  return game.fillFetch.hand;
};

const fetchRedCoin = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.redCoin) {
    let list = fetchObjsByType(game, OBJ_TYPES.RED_COIN);
    if (list.length !== 1) throw new Error('Multiple red coins found.'); // Something seriously went wrong
    game.fillFetch.redCoin = list[0];
  }
  return game.fillFetch.redCoin;
};

const fetchWhiteCoins = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.whiteCoins) {
    game.fillFetch.whiteCoins = [];
    let list = fetchObjsByType(game, OBJ_TYPES.WHITE_COIN);
    for (let obj of list)
      game.fillFetch.whiteCoins.push(obj);
  }
  return game.fillFetch.whiteCoins;
};

const fetchBlackCoins = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.blackCoins) {
    game.fillFetch.blackCoins = [];
    let list = fetchObjsByType(game, OBJ_TYPES.BLACK_COIN);
    for (let obj of list)
      game.fillFetch.blackCoins.push(obj);
  }
  return game.fillFetch.blackCoins;
};

const fetchPlayerColor = function(game, playerId) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.playerColors) game.fillFetch.playerColors = {};
  if (!game.fillFetch.playerColors[playerId]) {
    playerId = parseInt(playerId);
    game.fillFetch.playerColors[playerId] = game.getPlayerProp(playerId,
        ALL_PROP_KEYS.COLOR);
  }
  return game.fillFetch.playerColors[playerId];
};

const fetchOpponentId = function(game, selfPlayerId) {
  const gameKey = game.getId();
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.opponentMap) game.fillFetch.opponentMap = {};
  if (!game.fillFetch.opponentMap[selfPlayerId]) {
    let allPlayerIds = game.getAllPlayerIds();
    logger.trace({key: gameKey}, 'All player ids for fetching opponent ids are',
        allPlayerIds);
    if (allPlayerIds.length !== 2) return;
    game.fillFetch.opponentMap[allPlayerIds[0]] = allPlayerIds[1];
    game.fillFetch.opponentMap[allPlayerIds[1]] = allPlayerIds[0];
  }
  return game.fillFetch.opponentMap[selfPlayerId];
};

const fetchGameEndScore = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.gameEndScore)
    game.fillFetch.gameEndScore = fetchBlackCoins(game).length;

  return game.fillFetch.gameEndScore;
};

const fetchMeta = function(game) {
  if (!game.fillFetch) game.fillFetch = {};
  if (!game.fillFetch.gameMeta) game.fillFetch.gameMeta = game.getGameMeta();
  return game.fillFetch.gameMeta;
};

module.exports.genericFetchUtils = {
  fetchStriker,
  fetchHand,
  fetchRedCoin,
  fetchPlayerColor,
  fetchBlackCoins,
  fetchWhiteCoins,
  fetchOpponentId,
  fetchGameEndScore,
  fetchMeta,
  fetchObjById,
};