const logger = require('./playshipUtils').getLogger();
const UUID = require('uuid/v4');

const OBJ_TYPES = {
  RED_COIN: 'rcoin',
  BLACK_COIN: 'bcoin',
  WHITE_COIN: 'wcoin',
  STRIKER: 'striker',
  HOLE: 'hole',
  HAND: 'hand'
};

const COLOR = {
  RED: 'red',
  BLACK: 'black',
  WHITE: 'white',
};

module.exports.listOfTypes = function() {
  return OBJ_TYPES;
};

module.exports.allowedMargin = function() {
  return {
    ABANDON_MATCH: 7,
    ALL_OK: {
      x: 0.99,
      y: 0.01,
      z: 0.99
    },
  };
};

module.exports.getAllPropKeys = function() {
  return {
    IS_POTTED: 'isPotted',
    COLOR: 'color',
    POTT_TURN: 'pottTurn',
    POTT_TACK: 'pottTack',
    DID_SCORED_RED: 'didScoredRed',
    DID_POCKETED_RED: 'didPocketedRed',
    IS_SCORED: 'isScored',
    RECONCILE_STATUS: 'reconcileStatus',
    IS_ROTATION: 'isRotation',
    IDLE_TURN_COUNT: 'idleTurnCount',
    GRACE_TIME_LEFT_FOR_TURN: 'grTmFrTurn',
    TOTAL_GRACE_TIME_LEFT: 'ttlGrcTmLft',
    GRACE_TIME_START: 'graceTimeStart',
    IS_STRIKER_HIT: 'isStrikerHit',
    SHOULD_SEND_GAME_STATE: 'shouldSendGameState',
    IS_FORCE_TO_BE_REPLAYED: 'isFToBeRpld',
  };
};

module.exports.getAllTransientKeys = function() {
  return {
    TURN_RECONCILE_HOLDER: '_tc_holder',
    FORCE_REPLAY_COUNT: 'forceReplayCount',
  };
};

module.exports.getALLPropValues = function() {
  return {
    COLOR: COLOR,
    GRACE_TIME_GRANTED: 45 * 1000,
    GRACE_TIME_BUCKET_FOR_TURN: 20 * 1000,
    MIN_GRACE_TIME_TO_GRANT: 5 * 1000,
  };
};

module.exports.getColorOfCoinByType = function(type) {
  if (type === OBJ_TYPES.BLACK_COIN) return COLOR.BLACK;
  if (type === OBJ_TYPES.WHITE_COIN) return COLOR.WHITE;
  if (type === OBJ_TYPES.RED_COIN) return COLOR.RED;
};

module.exports.casesForSnapshotMatchCheck = function() {
  return {
    ALL_OK: 1,
    REVERT_TO_PREVIOUS_SAVEPOINT: 2,
    ABANDON_MATCH: 3,
  };
};

module.exports.getLogger = function() {
  return logger;
};

module.exports.parseBoolean = function(str, defValue) {
  if (!str) return defValue || false;
  return JSON.parse(str);
};

module.exports.FILL_FETCH_DEFAULT = UUID();

module.exports.getKOWinType = function() {
  return 2;
};

module.exports.getAbondonWinType = function() {
  return 0;
};

module.exports.getTurnCoolOffPeriod = function() {
  return 2500;
};

module.exports.getTurnReconCoolOffPeriod = function() {
  return 6000;
};

module.exports.getPreForceReplayCoolOffPeriod = function() {
  return 8000;
};

module.exports.getListOfCenteringBoundaries = function() {
  return [
    {
      minX: -0.136,
      minY: -0.136,
      maxX: 0.136,
      maxY: 0.136,
      posX: 0,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: 0.001799999999999996,
      minY: 0.10249999999999998,
      maxX: 0.27380000000000004,
      maxY: 0.3745,
      posX: 0.1378,
      posY: 0.135,
      posZ: 0.2385,
    },
    {
      minX: -0.27380000000000004,
      minY: -0.3745,
      maxX: -0.001799999999999996,
      maxY: -0.10249999999999998,
      posX: -0.1378,
      posY: 0.135,
      posZ: -0.2385,
    },
    {
      minX: -0.27380000000000004,
      minY: 0.10249999999999998,
      maxX: -0.001799999999999996,
      maxY: 0.3745,
      posX: -0.1378,
      posY: 0.135,
      posZ: 0.2385,
    },
    {
      minX: 0.001799999999999996,
      minY: -0.3745,
      maxX: 0.27380000000000004,
      maxY: -0.10249999999999998,
      posX: 0.1378,
      posY: 0.135,
      posZ: -0.2385,
    },
    {
      minX: 0.14,
      minY: -0.136,
      maxX: 0.41200000000000003,
      maxY: 0.136,
      posX: 0.276,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: -0.41200000000000003,
      minY: -0.136,
      maxX: -0.14,
      maxY: 0.136,
      posX: -0.276,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: 0.1395,
      minY: 0.3407,
      maxX: 0.41150000000000003,
      maxY: 0.6127,
      posX: 0.2755,
      posY: 0.135,
      posZ: 0.4767,
    },
    {
      minX: -0.41150000000000003,
      minY: -0.6127,
      maxX: -0.1395,
      maxY: -0.3407,
      posX: -0.2755,
      posY: 0.135,
      posZ: -0.4767,
    },
    {
      minX: -0.41150000000000003,
      minY: 0.3407,
      maxX: -0.1395,
      maxY: 0.6127,
      posX: -0.2755,
      posY: 0.135,
      posZ: 0.4767,
    },
    {
      minX: 0.1395,
      minY: -0.6127,
      maxX: 0.41150000000000003,
      maxY: -0.3407,
      posX: 0.2755,
      posY: 0.135,
      posZ: -0.4767,
    },
    {
      minX: -0.6880000000000001,
      minY: -0.136,
      maxX: -0.41600000000000004,
      maxY: 0.136,
      posX: -0.552,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: 0.41600000000000004,
      minY: -0.136,
      maxX: 0.6880000000000001,
      maxY: 0.136,
      posX: 0.552,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: -0.5495,
      minY: 0.10249999999999998,
      maxX: -0.27749999999999997,
      maxY: 0.3745,
      posX: -0.4135,
      posY: 0.135,
      posZ: 0.2385,
    },
    {
      minX: 0.27749999999999997,
      minY: -0.3745,
      maxX: 0.5495,
      maxY: -0.10249999999999998,
      posX: 0.4135,
      posY: 0.135,
      posZ: -0.2385,
    },
    {
      minX: -0.5495,
      minY: -0.3745,
      maxX: -0.27749999999999997,
      maxY: -0.10249999999999998,
      posX: -0.4135,
      posY: 0.135,
      posZ: -0.2385,
    },
    {
      minX: 0.27749999999999997,
      minY: 0.10249999999999998,
      maxX: 0.5495,
      maxY: 0.3745,
      posX: 0.4135,
      posY: 0.135,
      posZ: 0.2385,
    },
    {
      minX: -0.136,
      minY: 0.3407,
      maxX: 0.136,
      maxY: 0.6127,
      posX: 0,
      posY: 0.135,
      posZ: 0.4767,
    },
    {
      minX: -0.136,
      minY: -0.6127,
      maxX: 0.136,
      maxY: -0.3407,
      posX: 0,
      posY: 0.135,
      posZ: -0.4767,
    },
    {
      minX: 0.692,
      minY: -0.136,
      maxX: 0.964,
      maxY: 0.136,
      posX: 0.828,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: 0.5535,
      minY: 0.10249999999999998,
      maxX: 0.8255,
      maxY: 0.3745,
      posX: 0.6895,
      posY: 0.135,
      posZ: 0.2385,
    },
    {
      minX: 0.4155,
      minY: 0.3407,
      maxX: 0.6875,
      maxY: 0.6127,
      posX: 0.5515,
      posY: 0.135,
      posZ: 0.4767,
    },
    {
      minX: 0.27749999999999997,
      minY: 0.5787,
      maxX: 0.5495,
      maxY: 0.8507,
      posX: 0.4135,
      posY: 0.135,
      posZ: 0.7147,
    },
    {
      minX: 0.0020000000000000018,
      minY: 0.579,
      maxX: 0.274,
      maxY: 0.851,
      posX: 0.138,
      posY: 0.135,
      posZ: 0.715,
    },
    {
      minX: -0.274,
      minY: 0.579,
      maxX: -0.0020000000000000018,
      maxY: 0.851,
      posX: -0.138,
      posY: 0.135,
      posZ: 0.715,
    },
    {
      minX: -0.5495,
      minY: 0.5787,
      maxX: -0.27749999999999997,
      maxY: 0.8507,
      posX: -0.4135,
      posY: 0.135,
      posZ: 0.7147,
    },
    {
      minX: -0.6875,
      minY: 0.3407,
      maxX: -0.4155,
      maxY: 0.6127,
      posX: -0.5515,
      posY: 0.135,
      posZ: 0.4767,
    },
    {
      minX: -0.8255,
      minY: 0.10249999999999998,
      maxX: -0.5535,
      maxY: 0.3745,
      posX: -0.6895,
      posY: 0.135,
      posZ: 0.2385,
    },
    {
      minX: -0.964,
      minY: -0.136,
      maxX: -0.692,
      maxY: 0.136,
      posX: -0.828,
      posY: 0.135,
      posZ: 0,
    },
    {
      minX: -0.8255,
      minY: -0.3745,
      maxX: -0.5535,
      maxY: -0.10249999999999998,
      posX: -0.6895,
      posY: 0.135,
      posZ: -0.2385,
    },
    {
      minX: -0.6875,
      minY: -0.6127,
      maxX: -0.4155,
      maxY: -0.3407,
      posX: -0.5515,
      posY: 0.135,
      posZ: -0.4767,
    },
    {
      minX: -0.5495,
      minY: -0.8507,
      maxX: -0.27749999999999997,
      maxY: -0.5787,
      posX: -0.4135,
      posY: 0.135,
      posZ: -0.7147,
    },
    {
      minX: -0.274,
      minY: -0.851,
      maxX: -0.0020000000000000018,
      maxY: -0.579,
      posX: -0.138,
      posY: 0.135,
      posZ: -0.715,
    },
    {
      minX: 0.0020000000000000018,
      minY: -0.851,
      maxX: 0.274,
      maxY: -0.579,
      posX: 0.138,
      posY: 0.135,
      posZ: -0.715,
    },
    {
      minX: 0.27749999999999997,
      minY: -0.8507,
      maxX: 0.5495,
      maxY: -0.5787,
      posX: 0.4135,
      posY: 0.135,
      posZ: -0.7147,
    },
    {
      minX: 0.4155,
      minY: -0.6127,
      maxX: 0.6875,
      maxY: -0.3407,
      posX: 0.5515,
      posY: 0.135,
      posZ: -0.4767,
    },
    {
      minX: 0.5535,
      minY: -0.3745,
      maxX: 0.8255,
      maxY: -0.10249999999999998,
      posX: 0.6895,
      posY: 0.135,
      posZ: -0.2385,
    }];
};