const getPlayerProp = function(game, playerId, key) {
  if (typeof key === 'string')
    return game.getPlayerProp(playerId, key);
  let val = game.getPlayerProp(playerId, key.name);
  switch (key.type) {
    case 'boolean':
      return parseBool(val);
    case 'number':
      return parseFloat(val);
    case 'object':
      return JSON.parse(val || null);
    case 'objectSync':
      let o = JSON.parse(val || null);
      if (o) return o[key.name];
      else return o;
    case 'csv':
      if (val) return val.split(',');
      return val;
    default:
      return val;
  }
};

const setPlayerProp = function(game, playerId, key, value) {
  if (typeof key === 'string')
    return game.setPlayerProp(playerId, key, value);
  switch (key.type) {
    case 'object':
      if (value)
        return game.setPlayerProp(playerId, key.name, JSON.stringify(value));
      else
        return game.setPlayerProp(playerId, key.name, value);
    case 'objectSync':
      if (value)
        return game.setPlayerProp(playerId, key.name,
            JSON.stringify({[key.name]: value}));
      else
        return game.setPlayerProp(playerId, key.name, value);
    default:
      return game.setPlayerProp(playerId, key.name, value);
  }
};

const getGameProp = function(game, key) {
  if (typeof key === 'string')
    return game.getGameProp(key);
  let val = game.getGameProp(key.name);
  switch (key.type) {
    case 'boolean':
      return parseBool(val);
    case 'number':
      return parseFloat(val);
    case 'object':
      return JSON.parse(val || null);
    default:
      return val;
  }
};

const setGameProp = function(game, key, value) {
  if (typeof key === 'string')
    return game.setGameProp(key, value);
  switch (key.type) {
    case 'object':
      if (value)
        return game.setGameProp(key.name, JSON.stringify(value));
      else
        return game.setGameProp(key.name, value);
    default:
      return game.setGameProp(key.name, value);
  }
};


module.exports = {
  _getPlayerProp  : getPlayerProp,
  _setPlayerProp : setPlayerProp,
  _getGameProp : getGameProp,
  _setGameProp : setGameProp
}